@@ -28,15 +28,17 @@
 #include <util/sha1hash.h>
 #include <util/waitjob.h>
 #include <util/fileops.h>
+#include <util/functions.h>
 #include <torrent/globals.h>
 #include <torrent/torrent.h>
 #include <torrent/torrentcontrol.h>
+#include <torrent/jobqueue.h>
 #include <interfaces/torrentinterface.h>
 #include <interfaces/trackerslist.h>
 #include <settings.h>
-
 #include <climits>
 
+
 using namespace bt;
 
 namespace kt
@@ -55,7 +57,9 @@ namespace kt
 
 
 	QueueManager::~QueueManager()
-	{}
+	{
+		qDeleteAll(downloads);
+	}
 
 	void QueueManager::append(bt::TorrentInterface* tc)
 	{
@@ -63,25 +67,29 @@ namespace kt
 		downloads.sort();
 		connect(tc, SIGNAL(diskSpaceLow(bt::TorrentInterface*, bool)), this, SLOT(onLowDiskSpace(bt::TorrentInterface*, bool)));	
 		connect(tc, SIGNAL(torrentStopped(bt::TorrentInterface*)), this, SLOT(torrentStopped(bt::TorrentInterface*)));
+		connect(tc,SIGNAL(updateQueue()),this,SLOT(orderQueue()));
 	}
 
 	void QueueManager::remove(bt::TorrentInterface* tc)
 	{
 		paused_torrents.erase(tc);
 		int index = downloads.indexOf(tc);
-		if(index != -1)
-			delete downloads.takeAt(index);
+		if (index != -1)
+			downloads.takeAt(index)->deleteLater();
 	}
 
 	void QueueManager::clear()
 	{
+		exiting = true;
 		Uint32 nd = downloads.count();
-		downloads.clear();
 		paused_torrents.clear();
 		
 		// wait for a second to allow all http jobs to send the stopped event
 		if (nd > 0)
 			SynchronousWait(1000);
+		
+		qDeleteAll(downloads);
+		downloads.clear();
 	}
 	
 	TorrentStartResponse QueueManager::startInternal(bt::TorrentInterface* tc)
@@ -94,14 +102,10 @@ namespace kt
 			switch (Settings::startDownloadsOnLowDiskSpace())
 			{
 			case 0: //don't start!
-				tc->setPriority(0);
 				return bt::NOT_ENOUGH_DISKSPACE;
 			case 1: //ask user
 				if (KMessageBox::questionYesNo(0, i18n("You don't have enough disk space to download this torrent. Are you sure you want to continue?"), i18n("Insufficient disk space for %1",s.torrent_name)) == KMessageBox::No)
-				{
-					tc->setPriority(0);
 					return bt::USER_CANCELED;
-				}
 				else
 					break;
 			case 2: //force start
@@ -114,13 +118,16 @@ namespace kt
 		bool max_seed_time_reached = tc->overMaxSeedTime();
 		if (s.completed && (max_ratio_reached || max_seed_time_reached))
 		{
+			if (!enabled())
+				return QM_LIMITS_REACHED;
+			
 			QString msg; 
 			if (max_ratio_reached && max_seed_time_reached)
-				msg = msg = i18n("The torrent \"%1\" has reached it's maximum share ratio and it's maximum seed time. Ignore the limit and start seeding anyway?",s.torrent_name);
+				msg = i18n("The torrent \"%1\" has reached its maximum share ratio and its maximum seed time. Ignore the limit and start seeding anyway?",s.torrent_name);
 			else if (max_ratio_reached && !max_seed_time_reached)
-				msg = i18n("The torrent \"%1\" has reached it's maximum share ratio. Ignore the limit and start seeding anyway?",s.torrent_name);
+				msg = i18n("The torrent \"%1\" has reached its maximum share ratio. Ignore the limit and start seeding anyway?",s.torrent_name);
 			else if (max_seed_time_reached && !max_ratio_reached)
-				msg = msg = i18n("The torrent \"%1\" has reached it's maximum seed time. Ignore the limit and start seeding anyway?",s.torrent_name);
+				msg = i18n("The torrent \"%1\" has reached its maximum seed time. Ignore the limit and start seeding anyway?",s.torrent_name);
 				
 			if (KMessageBox::questionYesNo(0, msg, i18n("Maximum share ratio limit reached.")) == KMessageBox::Yes)
 			{
@@ -139,54 +146,77 @@ namespace kt
 		return START_OK;
 	}
 
-	TorrentStartResponse QueueManager::start(bt::TorrentInterface* tc, bool user)
+	TorrentStartResponse QueueManager::start(bt::TorrentInterface* tc)
 	{
-		const TorrentStats & s = tc->getStats();
-		bool start_tc = user;
-		bool check_done = false;
-
-		if (tc->isCheckingData(check_done) && !check_done)
-			return BUSY_WITH_DATA_CHECK;
-
-		if (!user)
-		{
-			if (s.completed)
-				start_tc = (max_seeds == 0 || getNumRunning(SEEDS) < max_seeds);
-			else
-				start_tc = (max_downloads == 0 || getNumRunning(DOWNLOADS) < max_downloads);
-		}
-		else
-		{
-			//User started this torrent so make it user controlled
-			tc->setPriority(0);
-		}
-
-		if (start_tc)
+		if (!enabled())
 		{
 			return startInternal(tc);
 		}
 		else
 		{
-			return QM_LIMITS_REACHED;
+			tc->setAllowedToStart(true);
+			bool start_tc = false;
+			if (tc->getJobQueue()->runningJobs())
+				return BUSY_WITH_JOB;
+			
+			const TorrentStats & s = tc->getStats();
+			if (!s.completed && !tc->checkDiskSpace(false)) //no need to check diskspace for seeding torrents
+			{
+				//we're short!
+				switch (Settings::startDownloadsOnLowDiskSpace())
+				{
+					case 0: //don't start!
+						return bt::NOT_ENOUGH_DISKSPACE;
+					case 1: //ask user
+						if (KMessageBox::questionYesNo(0, i18n("You don't have enough disk space to download this torrent. Are you sure you want to continue?"), i18n("Insufficient disk space for %1",s.torrent_name)) == KMessageBox::No)
+							return bt::USER_CANCELED;
+						else
+							break;
+					case 2: //force start
+						break;
+				}
+			}
+			
+			if (s.completed)
+				start_tc = (max_seeds == 0 || getNumRunning(SEEDS) < max_seeds);
+			else
+				start_tc = (max_downloads == 0 || getNumRunning(DOWNLOADS) < max_downloads);
+
+			if (start_tc)
+				orderQueue();
+			else
+				return QM_LIMITS_REACHED;
+			
+			return START_OK;
 		}
-		
-		return START_OK;
 	}
 
-	void QueueManager::stop(bt::TorrentInterface* tc, bool user)
+	void QueueManager::stop(bt::TorrentInterface* tc)
 	{
 		bool check_done = false;
-		if (tc->isCheckingData(check_done) && !check_done)
+		if (tc->getJobQueue()->runningJobs())
 			return;
 		
 		const TorrentStats & s = tc->getStats();
-		if (s.running)
-		{
-			stopSafely(tc,user);
-		}
+		if (enabled())
+			tc->setAllowedToStart(false);
 		
-		if (user) //dequeue it
-			tc->setPriority(0);
+		if (s.running)
+			stopSafely(tc);
+		else
+			tc->setQueued(false);
+	}
+	
+	void QueueManager::stop(QList<bt::TorrentInterface*> & todo)
+	{
+		ordering = true;
+		foreach (bt::TorrentInterface* tc,todo)
+		{
+			stop(tc);
+		}
+		ordering = false;
+		if (enabled())
+			orderQueue();
 	}
 	
 	void QueueManager::checkDiskSpace(QList<bt::TorrentInterface*> & todo)
@@ -238,7 +268,7 @@ namespace kt
 		foreach (bt::TorrentInterface* tc,todo)
 		{
 			const TorrentStats & s = tc->getStats();
-			if (tc->overMaxSeedTime())
+			if (s.completed && tc->overMaxSeedTime())
 			{
 				names.append(s.torrent_name);
 				tmp.append(tc);
@@ -267,7 +297,7 @@ namespace kt
 		foreach (bt::TorrentInterface* tc,todo)
 		{
 			const TorrentStats & s = tc->getStats();
-			if (tc->overMaxRatio())
+			if (s.completed && tc->overMaxRatio())
 			{
 				names.append(s.torrent_name);
 				tmp.append(tc);
@@ -314,68 +344,73 @@ namespace kt
 			if (s.running)
 				continue;
 			
-			bool check_done = false;
-			if (tc->isCheckingData(check_done) && !check_done)
+			if (tc->getJobQueue()->runningJobs())
 				continue;
 			
-			tc->setPriority(0);
-			startSafely(tc);
+			if (enabled())
+				tc->setAllowedToStart(true);
+			else
+				startSafely(tc);
 		}
+		
+		if (enabled())
+			orderQueue();
 	}
 	
-	void QueueManager::startall(int type)
+	void QueueManager::startAll()
 	{
+		if (enabled())
+		{
+			foreach (bt::TorrentInterface* tc,downloads)
+				tc->setAllowedToStart(true);
+			
+			orderQueue();
+		}
+		else
+		{
+			// first get the list of torrents which need to be started
+			QList<bt::TorrentInterface*> todo;
+			foreach (bt::TorrentInterface* tc,downloads)
+			{
+				const TorrentStats & s = tc->getStats();
+				if (s.running)
+					continue;
+				
+				if (tc->getJobQueue()->runningJobs())
+					continue;
+				
+				todo.append(tc);
+			}
+			
+			start(todo);
+		}
+	}
+
+	void QueueManager::stopAll()
+	{
+		stop(downloads);
+	}
+	
+	
+	void QueueManager::startAutoStartTorrents()
+	{
+		if (enabled()) 
+			return;
+		
 		// first get the list of torrents which need to be started
 		QList<bt::TorrentInterface*> todo;
 		foreach (bt::TorrentInterface* tc,downloads)
 		{
 			const TorrentStats & s = tc->getStats();
-			if (s.running)
+			if (s.running || tc->getJobQueue()->runningJobs() || !s.autostart)
 				continue;
 			
-			bool check_done = false;
-			if (tc->isCheckingData(check_done) && !check_done)
-				continue;
-			
-			if ((s.completed && type == 2) || (!s.completed && type == 1) || (type == 3))
-			{
-				todo.append(tc);
-			}
+			todo.append(tc);
 		}
 		
 		start(todo);
 	}
 
-	void QueueManager::stopall(int type)
-	{
-		QList<bt::TorrentInterface *>::iterator i = downloads.begin();
-		while (i != downloads.end())
-		{
-			bt::TorrentInterface* tc = *i;
-			const TorrentStats & s = tc->getStats();
-			if (tc->getStats().running)
-			{
-				try
-				{
-					if(type >= 3)
-						stopSafely(tc,true);
-					else if( (s.completed && type == 2) || (!s.completed && type == 1) )
-						stopSafely(tc,true);
-				}
-				catch (bt::Error & err)
-				{
-					QString msg =
-							i18n("Error stopping torrent %1 : %2",
-							s.torrent_name,err.toString());
-					KMessageBox::error(0,msg,i18n("Error"));
-				}
-			}
-			else //if torrent is not running but it is queued we need to make it user controlled
-				if( (s.completed && type == 2) || (!s.completed && type == 1) || (type == 3) )
-					tc->setPriority(0); 
-			i++;
-		}
-	}
 	
 	void QueueManager::onExit(WaitJob* wjob)
 	{
@@ -386,7 +421,7 @@ namespace kt
 			bt::TorrentInterface* tc = *i;
 			if (tc->getStats().running)
 			{
-				stopSafely(tc,false,wjob);
+				stopSafely(tc,wjob);
 			}
 			i++;
 		}
@@ -410,8 +445,8 @@ namespace kt
 	int QueueManager::getNumRunning(Flags flags)
 	{
 		int nr = 0;
-		QList<TorrentInterface*>::const_iterator i = downloads.begin();
-		while (i != downloads.end())
+		QList<TorrentInterface*>::const_iterator i = downloads.constBegin();
+		while (i != downloads.constEnd())
 		{
 			const TorrentInterface* tc = *i;
 			const TorrentStats & s = tc->getStats();
@@ -459,9 +494,14 @@ namespace kt
 	
 	void QueueManager::onLowDiskSpace(bt::TorrentInterface* tc, bool toStop)
 	{
-		if(toStop)
+		if (toStop)
 		{
-			stop(tc, false);
+			stopSafely(tc);
+			if (enabled())
+			{
+				tc->setAllowedToStart(false);
+				orderQueue();
+			}
 		}
 		
 		//then emit the signal to inform trayicon to show passive popup
@@ -503,25 +543,30 @@ namespace kt
 	
 	void QueueManager::orderQueue()
 	{
-		if (ordering || !downloads.count() || paused_state || exiting)
+		if (ordering || !downloads.count() || exiting)
+			return;
+		
+		downloads.sort(); // sort downloads, even when paused so that the QM widget is updated
+		if (Settings::manuallyControlTorrents() || paused_state)
 			return;
 		
 		ordering = true; // make sure that recursive entering of this function is not possible
 		
-		downloads.sort();
-	
 		QueuePtrList download_queue;
 		QueuePtrList seed_queue;
 		
 		foreach (TorrentInterface* tc,downloads)
 		{
 			const TorrentStats & s = tc->getStats();
-			
-			if (!s.user_controlled && !tc->isMovingFiles() && !s.stopped_by_error)
+			bool dummy;
+			if (tc->isAllowedToStart() && !s.stopped_by_error && !tc->getJobQueue()->runningJobs())
 			{
 				if (s.completed)
-					seed_queue.append(tc);
-				else		
+				{
+					if (!tc->overMaxRatio() && !tc->overMaxSeedTime())
+						seed_queue.append(tc);
+				}
+				else
 					download_queue.append(tc);
 			}
 		}
@@ -538,8 +583,6 @@ namespace kt
 					Out(SYS_GEN|LOG_DEBUG) << "QM Starting: " << s.torrent_name << endl;
 					if (startInternal(tc) == bt::START_OK)
 						num_running++;
-					else if (s.stopped_by_error)
-						tc->setPriority(0);
 				}
 				else
 					num_running++;
@@ -549,8 +592,9 @@ namespace kt
 				if (s.running)
 				{
 					Out(SYS_GEN|LOG_DEBUG) << "QM Stopping: " << s.torrent_name << endl;
-					stop(tc);
+					stopSafely(tc);
 				}
+				tc->setQueued(true);
 			}
 		}
 		
@@ -558,7 +602,6 @@ namespace kt
 		foreach (bt::TorrentInterface* tc,seed_queue)
 		{
 			const TorrentStats & s = tc->getStats();
-
 			if (num_running < max_seeds || max_seeds == 0)
 			{
 				if (!s.running)
@@ -566,8 +609,6 @@ namespace kt
 					Out(SYS_GEN|LOG_DEBUG) << "QM Starting: " << s.torrent_name << endl;
 					if (startInternal(tc) == bt::START_OK)
 						num_running++;
-					else if (s.stopped_by_error)
-						tc->setPriority(0);
 				}
 				else
 					num_running++;
@@ -577,8 +618,9 @@ namespace kt
 				if (s.running)
 				{
 					Out(SYS_GEN|LOG_DEBUG) << "QM Stopping: " << s.torrent_name << endl;
-					stop(tc);
+					stopSafely(tc);
 				}
+				tc->setQueued(true);
 			}
 		}
 		
@@ -588,39 +630,37 @@ namespace kt
 	
 	void QueueManager::torrentFinished(bt::TorrentInterface* tc)
 	{
-		//dequeue this tc
-		tc->setPriority(0);
-		//make sure the max_seeds is not reached
-// 		if(max_seeds !=0 && max_seeds < getNumRunning(false,true))
-// 			tc->stop(true);
-		
-		if (keep_seeding)
-			torrentAdded(tc,false,false);
-		else
-			stopSafely(tc,true);
+		if (!keep_seeding)
+		{
+			if (enabled())
+				tc->setAllowedToStart(false);
+			
+			stopSafely(tc);
+		}
 		
 		orderQueue();
 	}
 	
-	void QueueManager::torrentAdded(bt::TorrentInterface* tc,bool user, bool start_torrent)
+	void QueueManager::torrentAdded(bt::TorrentInterface* tc,bool start_torrent)
 	{
-		if (!user)
+		if (enabled())
 		{
+			// new torrents have the lowest priority
+			// so everybody else gets a higher priority
 			foreach (TorrentInterface * otc,downloads)
 			{
 				int p = otc->getPriority();
-				if (p > 0)
-					otc->setPriority(p+1);
+				otc->setPriority(p+1);
 			}
-			tc->setPriority(1);
+			tc->setAllowedToStart(start_torrent);
+			tc->setPriority(0);
 			rearrangeQueue();
 			orderQueue();
 		}
 		else
 		{
-			 tc->setPriority(0);
-			 if(start_torrent)
-				 start(tc, true);
+			 if (start_torrent)
+				 start(tc);
 		}	
 	}
 	
@@ -633,9 +673,13 @@ namespace kt
 	
 	void QueueManager::setPausedState(bool pause)
 	{
+		if (paused_state == pause)
+			return;
+		
 		paused_state = pause;	
 		if(!pause)
 		{
+			UpdateCurrentTime();
 			std::set<bt::TorrentInterface*>::iterator it = paused_torrents.begin();
 			while (it != paused_torrents.end())
 			{
@@ -651,8 +695,8 @@ namespace kt
 		{
 			foreach (TorrentInterface* tc,downloads)
 			{
-				const TorrentStats & s = tc->getStats();         
-				if(s.running)
+				const TorrentStats & s = tc->getStats();
+				if (s.running)
 				{
 					paused_torrents.insert(tc);
 					stopSafely(tc,false);
@@ -669,39 +713,10 @@ namespace kt
 		// make sure everybody has an unique priority
 		foreach (bt::TorrentInterface* tc,downloads)
 		{
-			if (tc->getPriority() > 0)
-				tc->setPriority(prio--);
+			tc->setPriority(prio--);
 		}
 	}
 	
-	void QueueManager::enqueue(bt::TorrentInterface* tc)
-	{
-		//if a seeding torrent reached its maximum share ratio or maximum seed time don't enqueue it...
-		if (tc->getStats().completed && (tc->overMaxRatio() || tc->overMaxSeedTime()))
-		{
-			Out(SYS_GEN | LOG_IMPORTANT) << "Torrent has reached max share ratio or max seed time and cannot be started automatically." << endl;
-			emit queuingNotPossible(tc);
-			return;
-		}
-		
-		torrentAdded(tc,false,false);
-	}
-	
-	void QueueManager::dequeue(bt::TorrentInterface* tc)
-	{
-		tc->setPriority(0);
-		rearrangeQueue();
-		orderQueue();
-	}
-	
-	void QueueManager::queue(bt::TorrentInterface* tc)
-	{
-		if(tc->getPriority() == 0)
-			enqueue(tc);
-		else
-			dequeue(tc);
-	}
-	
 	void QueueManager::startSafely(bt::TorrentInterface* tc)
 	{
 		try
@@ -718,11 +733,11 @@ namespace kt
 		}
 	}
 	
-	void QueueManager::stopSafely(bt::TorrentInterface* tc,bool user,WaitJob* wjob)
+	void QueueManager::stopSafely(bt::TorrentInterface* tc,WaitJob* wjob)
 	{
 		try
 		{
-			tc->stop(user,wjob);
+			tc->stop(wjob);
 		}
 		catch (bt::Error & err)
 		{
@@ -752,34 +767,44 @@ namespace kt
 	
 	void QueueManager::checkStalledTorrents(bt::TimeStamp now,bt::Uint32 min_stall_time)
 	{
-		const int MAX_PRIO = INT_MAX;
-		bool found_stalled = false;
-		// set the priority of all stalled ones to -1
+		if (!enabled())
+			return;
+		
+		QueuePtrList newlist;
+		QueuePtrList stalled;
+		bool can_decrease = false;
+		
+		// find all stalled ones
 		foreach (bt::TorrentInterface* tc,downloads)
 		{
-			if (tc->getPriority() > 1 && IsStalled(tc,now,min_stall_time))
+			if (IsStalled(tc,now,min_stall_time))
 			{
-				tc->setPriority(MAX_PRIO);
-				found_stalled = true;
-				Out(SYS_GEN | LOG_NOTICE) << "The torrent " << tc->getStats().torrent_name << " has stalled longer then " << min_stall_time << " minutes, decreasing it's priority" << endl;
+				stalled.append(tc);
+			}
+			else
+			{
+				// decreasing makes only sense if there are QM torrents after the stalled ones
+				can_decrease = stalled.count() > 0;
+				newlist.append(tc);
 			}
 		}
+		
+		if (stalled.count() == 0 || stalled.count() == downloads.count() || !can_decrease)
+			return;
+		
+		foreach (bt::TorrentInterface* tc,stalled)
+			Out(SYS_GEN | LOG_NOTICE) << "The torrent " << tc->getStats().torrent_name << " has stalled longer then " << min_stall_time << " minutes, decreasing it's priority" << endl;
 	
-		if (found_stalled)
+		downloads.clear();
+		downloads += newlist;
+		downloads += stalled;
+		// redo priorities and then order the queue
+		int prio = downloads.count();
+		foreach (bt::TorrentInterface* tc,downloads)
 		{
-			// redo the priority, stalled ones now get 1, 
-			// the other ones get an increase if they are not user controlled
-			foreach (bt::TorrentInterface* tc,downloads)
-			{
-				int prio = tc->getPriority();
-				if (prio == MAX_PRIO)
-					tc->setPriority(1);
-				else if (prio > 1 || (prio == 1 && !IsStalled(tc,now,min_stall_time)))
-					tc->setPriority(prio + 1);
-			}
-			
-			orderQueue();
+			tc->setPriority(prio--);
 		}
+		orderQueue();
 	}
 	/////////////////////////////////////////////////////////////////////////////////////////////
 
